/*
    -- MAGMA (version 1.2.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       June 2012

       @generated c Thu May  2 22:50:24 2013

*/
#include "common_magma.h"


extern "C" magma_int_t
magma_cpotrf_mmic(int num_mics, magma_uplo_t uplo, magma_int_t n, 
                 magmaFloatComplex **d_lA, magma_int_t ldda, magma_int_t *info, magma_queue_t queue )
{
/*  -- MAGMA (version 1.2.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       June 2012

    Purpose   
    =======   
    CPOTRF computes the Cholesky factorization of a complex Hermitian   
    positive definite matrix dA.   

    The factorization has the form   
       dA = U**H * U,  if UPLO = 'U', or   
       dA = L  * L**H,  if UPLO = 'L',   
    where U is an upper triangular matrix and L is lower triangular.   

    This is the block version of the algorithm, calling Level 3 BLAS.   

    Arguments   
    =========   
    UPLO    (input) CHARACTER*1   
            = 'U':  Upper triangle of dA is stored;   
            = 'L':  Lower triangle of dA is stored.   

    N       (input) INTEGER   
            The order of the matrix dA.  N >= 0.   

    dA      (input/output) COMPLEX array on the mic, dimension (LDDA,N)   
            On entry, the Hermitian matrix dA.  If UPLO = 'U', the leading   
            N-by-N upper triangular part of dA contains the upper   
            triangular part of the matrix dA, and the strictly lower   
            triangular part of dA is not referenced.  If UPLO = 'L', the   
            leading N-by-N lower triangular part of dA contains the lower   
            triangular part of the matrix dA, and the strictly upper   
            triangular part of dA is not referenced.   

            On exit, if INFO = 0, the factor U or L from the Cholesky   
            factorization dA = U**H * U or dA = L * L**H.   

    LDDA     (input) INTEGER   
            The leading dimension of the array dA.  LDDA >= max(1,N).
            To benefit from coalescent memory accesses LDDA must be
            dividable by 16.

    INFO    (output) INTEGER   
            = 0:  successful exit   
            < 0:  if INFO = -i, the i-th argument had an illegal value   
            > 0:  if INFO = i, the leading minor of order i is not   
                  positive definite, and the factorization could not be   
                  completed.   
    =====================================================================   */

	#define dlA(id, i, j)  (d_lA[id] + (j) * ldda + (i))
	#define dlP(id, i, j)  (d_lP[id] + (j) * ldda + (i))

	#define panel(j)  (panel + (j))
	#define tmppanel(j)  (tmppanel + (j))
	#define tmpprevpanel(j)  (tmpprevpanel + (j))

	magma_int_t     j, nb, d, id, j_local, j_local2, blkid, crosspoint, prevj, prevtrsmrows;
	magmaFloatComplex c_one     = MAGMA_C_ONE;
        magmaFloatComplex  z_one = MAGMA_C_MAKE(  1.0, 0.0 );
        magmaFloatComplex  mz_one = MAGMA_C_MAKE( -1.0, 0.0 );
	magmaFloatComplex c_neg_one = MAGMA_C_NEG_ONE;
	magmaFloatComplex *panel, *tmppanel0, *tmppanel1, *tmppanel, *tmpprevpanel;
	float          d_one     =  1.0;
	float          d_neg_one = -1.0;
	magmaFloatComplex *d_lP[MagmaMaxDevices], *dlpanel;
	magma_int_t rows, trsmrows, myrows, n_id, nmic, n_local[MagmaMaxDevices], lddat_local[MagmaMaxDevices], lddat, ldpanel;
	magma_event_t event[MagmaMaxDevices][4];

        float    one =  1.0;
        float  m_one = -1.0;

	*info = 0;
	if ( uplo != MagmaUpper && uplo != MagmaLower ) {
		*info = -1;
	} else if (n < 0) {
		*info = -2;
	} else if (ldda < max(1,n)) {
		*info = -4;
	}
	if (*info != 0) {
		magma_xerbla( __func__, -(*info) );
		return *info;
	}

	nb = magma_get_cpotrf_nb(n);

	ldpanel = ldda;
	if (MAGMA_SUCCESS != magma_cmalloc_pinned( &panel, 2 * nb * ldpanel )) {
		*info = MAGMA_ERR_HOST_ALLOC;
		return *info;
	}

	tmppanel0 = panel;
	tmppanel1 = tmppanel0 + nb * ldpanel;

	if ((nb <= 1) || (nb >= n)) {
		//Use unblocked code.
		magma_setdevice(0);
		magma_cgetmatrix( n, n, dlA(0, 0, 0), 0, ldda, panel, 0, ldpanel, queue);
		lapackf77_cpotrf( lapack_const(uplo), &n, panel, &ldpanel, info);
		magma_csetmatrix( n, n, panel, 0, ldpanel, dlA(0, 0, 0), 0, ldda, queue );
	} else {

		for( d = 0; d < num_mics; d++ ) {
			magma_setdevice(d);

			// local-n and local-ld 
			n_local[d] = ((n / nb) / num_mics) * nb;
			if (d < (n / nb) % num_mics)
				n_local[d] += nb;
			else if (d == (n / nb) % num_mics)
				n_local[d] += n % nb;

			lddat_local[d] = ((n_local[d] + 31) / 32) * 32;
			lddat_local[d] = (lddat_local[d] % 256 == 0) ? lddat_local[d] + 32 : lddat_local[d];

			if (MAGMA_SUCCESS != magma_cmalloc( &d_lP[d], nb * ldda )) {
				for( j = 0; j < d; j++ ) {
					magma_setdevice(j);
					magma_free( d_lP[d] );
				}
				*info = MAGMA_ERR_DEVICE_ALLOC;
				return *info;
			}

			for( j = 0; j < 4; j++ ) {
				event[d][j] = NULL;
			}
		}

		//#define ENABLE_TIMER
		#if defined (ENABLE_TIMER)
		float therk[4], tmtc, tcchol, tctrsm, tctm, tmnp, tcnp;
		printf("\n\n %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s %10s\n", 
				"j", "nb", "row", "mtc", "CPU_np", "panel", "ctrsm", "CH+TRSM", "CPU", "dsyrk[0]", "dsyrk[1]", "dsyrk[2]", "dsyrk[3]", "ctm P", "MIC_np");
		printf("     ====================================================================================================\n");
                #endif 

		//Use blocked code.
		if (uplo == MagmaUpper) 
		{    
		} else {

			blkid = -1;
			crosspoint = 0;//n; //n -- > mic always does next panel, 0 --> cpu always does next panel
			if(num_mics == 4)
				crosspoint = n;
			else if(num_mics == 3)
				crosspoint = n;
			else if(num_mics == 2)
				crosspoint = 20160;
			else
				crosspoint = 0;

			// Compute the Cholesky factorization A = L*L' 
			for (j = 0; (j + nb) < n; j += nb) {

                        	#if defined (ENABLE_TIMER)
                        	therk[0] = therk[1] = therk[2] = therk[3] = tmtc = tcchol = tctrsm = tctm = tmnp = tcnp = 0.0;
                        	#endif

				blkid += 1;
				tmppanel = (blkid % 2 == 0) ? tmppanel0 : tmppanel1;
				// Set the mic number that holds the current panel 
				id = (j / nb) % num_mics;
				magma_setdevice(id);

				//Set the local index where the current panel is
				j_local = j / (nb * num_mics) * nb;
				
				rows = n - j;
				//factor the diagonal
				magma_setdevice(id);

				if(j == 0 || prevtrsmrows <= crosspoint) {
                                	#if defined (ENABLE_TIMER)
					tmtc = get_time();
                                	#endif

					// send the diagonal to cpu 
					magma_cgetmatrix_async( rows, nb,
							dlA(id, j, j_local), 0, ldda,
							tmppanel(j), 0, ldpanel, queue, &event[id][0] );
				
					magma_event_sync( event[id][0] );

                                	#if defined (ENABLE_TIMER)
					tmtc = get_time() - tmtc;
                                	#endif
				}else {

					#if defined (ENABLE_TIMER)
					tcnp = get_time();
					#endif

					tmpprevpanel = ((blkid - 1) % 2) == 0 ? tmppanel0 : tmppanel1;
					//start sending the panel to cpu
					magma_cgetmatrix_async(rows, nb,
							dlA(id, j, j_local), 0, ldda,
							tmppanel(j),      0, ldpanel, queue, &event[id][0] );

					//stnchrnoize i-th panel from id-th mic into work
					magma_event_sync(event[id][0] );

					blasf77_cgemm( MagmaNoTransStr, MagmaConjTransStr,
							&rows, &nb, &nb,
							&mz_one, tmpprevpanel(j), &ldpanel,
							tmpprevpanel(j), &ldpanel,
							&z_one,     tmppanel(j), &ldpanel );

					#if defined (ENABLE_TIMER)
					tcnp = get_time() - tcnp;
					#endif
				}



                                #if defined (ENABLE_TIMER)
                                tcchol = get_time();
                                #endif

				lapackf77_cpotrf(MagmaLowerStr, &nb, tmppanel(j), &ldpanel, info);
				if (*info != 0) {
					*info = *info + j;
					//break;
				}

                                #if defined (ENABLE_TIMER)
                                tcchol = get_time() - tcchol;
                                #endif

                                #if defined (ENABLE_TIMER)
                                tctrsm = get_time();
                                #endif

				trsmrows = rows - nb;

				if(trsmrows > 0) {
					blasf77_ctrsm(MagmaRightStr, MagmaLowerStr, MagmaConjTransStr, MagmaNonUnitStr, &trsmrows, &nb, &z_one, tmppanel(j), &ldpanel, tmppanel(j + nb), &ldpanel);
				}

                                #if defined (ENABLE_TIMER)
                                tctrsm = get_time() - tctrsm;
                                #endif

                                #if defined (ENABLE_TIMER)
                                tctm = get_time();
                                #endif

				//send the diagonal to mics
				d = (id + 1) % num_mics;
				for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
					magma_setdevice(d);
					magma_int_t myrows = 0;
					magma_int_t row_offset = 0;
					if( d == id ) {
						dlpanel = dlA(d, j, j_local);
						myrows = rows;
						row_offset = 0;
					} else {
						dlpanel = dlP(d, 0, 0);
						myrows = trsmrows;
						row_offset = nb;
					}

					if(myrows > 0) {
						magma_kernel_sync();
						magma_csetmatrix_async(myrows, nb,
								tmppanel(j + row_offset),    0, ldpanel,
								dlpanel, 0, ldda, queue, &event[d][1] );
					}
				}

                                #if defined (ENABLE_TIMER)
				d = (id + 1) % num_mics;
				for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
					magma_setdevice(d);
					magma_event_sync( event[d][1] );
				}

				tctm = get_time() - tctm;
                                #endif

				if ( (j + nb) < n) {
					magma_int_t offset = 0;
					magma_int_t row_offset = 0;
					//do gemm on look ahead panel
					if(j + nb + nb < n) {
						if(trsmrows <= crosspoint) {
                                			#if defined (ENABLE_TIMER)
							tmnp = get_time();
                                			#endif

							magma_int_t la_id = (id + 1) % num_mics;
							magma_int_t j_local2 = (j + nb) / (nb * num_mics) * nb;
							magma_setdevice(la_id);
							magma_event_sync( event[la_id][1] );

							if( la_id == id ) {
								dlpanel = dlA(d, j + nb, j_local);
							} else {
								dlpanel = dlP(d, 0, 0);
							}

							magma_cgemm( MagmaNoTrans, MagmaConjTrans,
									trsmrows, nb, nb, 
									m_one, dlpanel, 0, ldda, dlpanel, 0, ldda, 
									one, dlA(la_id, j + nb, j_local2), 0, ldda, queue);


                                			#if defined (ENABLE_TIMER)
							magma_kernel_sync();
							tmnp = get_time() - tmnp;
                                			#endif
						}

						offset = j + nb + nb;
						row_offset = nb;

					} else {
						offset = j + nb;
						row_offset = 0;
					}

					//syrk on multiple mic
					d = (id + 1) % num_mics;
					for(nmic = 0; nmic < num_mics; nmic++, d = (d + 1) % num_mics ) {
						magma_setdevice(d);
						magma_event_sync( event[d][1] );
						magma_kernel_sync();

						//update the entire trailing matrix
						if( d == id ) {
							dlpanel = dlA(d, j + nb + row_offset, j_local);
						} else {
							dlpanel = dlP(d, row_offset, 0);
						}

                                		#if defined (ENABLE_TIMER)
						therk[d] = get_time();
                                		#endif

						if(n - offset > 0) {
							magma_cherk_mmic(MagmaLower, MagmaTrans, n - offset, nb,
									m_one, dlpanel, ldda, 0,
									one, d_lA[d], ldda,  offset,
									num_mics, nb, d);
						}

                                		#if defined (ENABLE_TIMER)
						magma_kernel_sync();
						therk[d] = get_time() - therk[d];
                                		#endif
					}

					prevtrsmrows = trsmrows;
					prevj = j;

					#if defined (ENABLE_TIMER)
					printf("%10d %10d %10d %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf %10.3lf(%d)\n",
							j, nb, rows, tmtc * 1000, tcnp*1000, tcchol*1000, tctrsm * 1000, (tcchol + tctrsm) * 1000, 
							(tmtc+tcnp+tcchol+tctrsm)*1000, therk[0] * 1000, therk[1] * 1000, therk[2] * 1000, therk[3] * 1000, tctm * 1000,  
							tmnp * 1000, (id + 1) % num_mics);
					fflush(0);
					#endif
	

				}
			}

			//cholesky for the last block
			if(j < n) {
				rows = n - j;
				id = (j / nb) % num_mics;
				magma_setdevice(id);
				magma_kernel_sync();

				//Set the local index where the current panel is
				j_local = j / (nb * num_mics) * nb;
				
				magma_cgetmatrix(rows, rows, dlA(id, j, j_local), 0, ldda, panel(j), 0, ldpanel, queue);
				lapackf77_cpotrf(MagmaLowerStr, &rows, panel(j), &ldpanel, info);
				magma_csetmatrix(rows, rows, panel(j), 0, ldpanel, dlA(id, j, j_local), 0, ldpanel, queue);
			}

		} // end of else not upper

		//clean up
		for( d = 0; d < num_mics; d++ ) {
			magma_setdevice(d);
			magma_free( d_lP[d] );
		}

	} // end of not lapack

	// free workspace
	magma_free_pinned( panel );

	return *info;
} /* magma_cpotrf_mmic */
