/*
 *  -- micMAGMA (version 0.3) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     April 2012
 *
 * @generated s Thu May  2 22:59:07 2013
 *
 **/
// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "testings.h"


// Initialize matrix to random.
// Having this in separate function ensures the same ISEED is always used,
// so we can re-generate the identical matrix.
void init_matrix( int m, int n, float *h_A, magma_int_t lda )
{
    magma_int_t ione = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    magma_int_t n2 = lda*n;
    lapackf77_slarnv( &ione, ISEED, &n2, h_A );
}


// On input, A and ipiv is LU factorization of A. On output, A is overwritten.
// Requires m == n.
// Uses init_matrix() to re-generate original A as needed.
// Generates random RHS b and solves Ax=b.
// Returns residual, |Ax - b| / (n |A| |x|).
float get_residual(
    magma_int_t m, magma_int_t n,
    float *A, magma_int_t lda,
    magma_int_t *ipiv )
{
    if ( m != n ) {
        printf( "\nERROR: residual check defined only for square matrices\n" );
        return -1;
    }
    
    float c_one     = MAGMA_S_ONE;
    float c_neg_one = MAGMA_S_NEG_ONE;
    magma_int_t ione = 1;
    
    // this seed should be DIFFERENT than used in init_matrix
    // (else x is column of A, so residual can be exactly zero)
    magma_int_t ISEED[4] = {0,0,0,2};
    magma_int_t info = 0;
    float *x, *b;
    
    // initialize RHS
    TESTING_MALLOC( x, float, n );
    TESTING_MALLOC( b, float, n );
    lapackf77_slarnv( &ione, ISEED, &n, b );
    blasf77_scopy( &n, b, &ione, x, &ione );
        
    // solve Ax = b
    lapackf77_sgetrs( "Notrans", &n, &ione, A, &lda, ipiv, x, &n, &info );
    if ( info != 0 )
        printf( "lapackf77_sgetrs returned error %d\n", info );
    
    // reset to original A
    init_matrix( m, n, A, lda );
    
    // compute r = Ax - b, saved in b
    blasf77_sgemv( "Notrans", &m, &n, &c_one, A, &lda, x, &ione, &c_neg_one, b, &ione );
    
    // compute residual |Ax - b| / (n*|A|*|x|)
    float norm_x, norm_A, norm_r, work[1];
    norm_A = lapackf77_slange( "F", &m, &n, A, &lda, work );
    norm_r = lapackf77_slange( "F", &n, &ione, b, &n, work );
    norm_x = lapackf77_slange( "F", &n, &ione, x, &n, work );
    
    //printf( "r=\n" ); magma_sprint( 1, n, b, 1 );
    
    TESTING_FREE( x );
    TESTING_FREE( b );
    
    //printf( "r=%.2e, A=%.2e, x=%.2e, n=%d\n", norm_r, norm_A, norm_x, n );
    return norm_r / (n * norm_A * norm_x);
}


// On input, LU and ipiv is LU factorization of A. On output, LU is overwritten.
// Works for any m, n.
// Uses init_matrix() to re-generate original A as needed.
// Returns error in factorization, |PA - LU| / (n |A|)
// This allocates 3 more matrices to store A, L, and U.
float get_LU_error(magma_int_t M, magma_int_t N, 
                    float *LU, magma_int_t lda,
                    magma_int_t *ipiv)
{
    magma_int_t min_mn = min(M,N);
    magma_int_t ione   = 1;
    magma_int_t i, j;
    float alpha = MAGMA_S_ONE;
    float beta  = MAGMA_S_ZERO;
    float *A, *L, *U;
    float work[1], matnorm, residual;
    
    TESTING_MALLOC( A, float, lda*N    );
    TESTING_MALLOC( L, float, M*min_mn );
    TESTING_MALLOC( U, float, min_mn*N );
    memset( L, 0, M*min_mn*sizeof(float) );
    memset( U, 0, min_mn*N*sizeof(float) );

    // set to original A
    init_matrix( M, N, A, lda );
    lapackf77_slaswp( &N, A, &lda, &ione, &min_mn, ipiv, &ione);
    
    // copy LU to L and U, and set diagonal to 1
    lapackf77_slacpy( MagmaLowerStr, &M, &min_mn, LU, &lda, L, &M      );
    lapackf77_slacpy( MagmaUpperStr, &min_mn, &N, LU, &lda, U, &min_mn );
    for(j=0; j<min_mn; j++)
        L[j+j*M] = MAGMA_S_MAKE( 1., 0. );
    
    matnorm = lapackf77_slange("f", &M, &N, A, &lda, work);

    blasf77_sgemm("N", "N", &M, &N, &min_mn,
                  &alpha, L, &M, U, &min_mn, &beta, LU, &lda);

    for( j = 0; j < N; j++ ) {
        for( i = 0; i < M; i++ ) {
            LU[i+j*lda] = MAGMA_S_SUB( LU[i+j*lda], A[i+j*lda] );
        }
    }
    residual = lapackf77_slange("f", &M, &N, LU, &lda, work);

    TESTING_FREE(A);
    TESTING_FREE(L);
    TESTING_FREE(U);

    return residual / (matnorm * N);
}

#define PRECISION_s


/* ////////////////////////////////////////////////////////////////////////////
   -- Testing sgetrf
*/
int main( int argc, char** argv)
{
    real_Double_t   gflops, gpu_perf, gpu_time, cpu_perf, cpu_time;
    float          error;
    float *h_A;
    magma_int_t     *ipiv;

    /* Matrix size */
    magma_int_t M = 0, N = 0, n2, lda;
    const int MAXTESTS = 10;
    magma_int_t msize[MAXTESTS] = { 1024, 2048, 3072, 4032, 5184, 6016, 7040, 8064, 9088, 10112 };
    magma_int_t nsize[MAXTESTS] = { 1024, 2048, 3072, 4032, 5184, 6016, 7040, 8064, 9088, 10112 };

    magma_int_t i, info, min_mn, nb, ret;

    int lapack   = getenv("MAGMA_RUN_LAPCAK")     != NULL;
    int checkres = getenv("MAGMA_TESTINGS_CHECK") != NULL;
    
    // process command line arguments
    printf( "\nUsage: %s -N <m,n> -c -c2 -l\n"
            "  -N  can be repeated up to %d times. If only m is given, then m=n.\n"
            "  -c  or setting $MAGMA_TESTINGS_CHECK checks result, PA - LU.\n"
            "  -c2 for square matrices, solves one RHS and checks residual, Ax - b.\n"
            "  -l  or setting $MAGMA_RUN_LAPACK runs LAPACK.\n\n",
            argv[0], MAXTESTS );
    
    int ntest = 0;
    for( int i = 1; i < argc; ++i ) {
        if ( strcmp("-N", argv[i]) == 0 && i+1 < argc ) {
	    if (ntest >= MAXTESTS) {
	        printf("error: -N repeated more than maximum %d tests\n", MAXTESTS );
                exit(1);
            }
            int m, n;
            info = sscanf( argv[++i], "%d,%d", &m, &n );
            if ( info == 2 && m > 0 && n > 0 ) {
                msize[ ntest ] = m;
                nsize[ ntest ] = n;
            }
            else if ( info == 1 && m > 0 ) {
                msize[ ntest ] = m;
                nsize[ ntest ] = m;  // implicitly
            }
            else {
                printf( "error: -N %s is invalid; ensure m > 0, n > 0.\n", argv[i] );
                exit(1);
            }
            M = max( M, msize[ ntest ] );
            N = max( N, nsize[ ntest ] );
            ntest++;
        }
        else if ( strcmp("-M", argv[i]) == 0 ) {
            printf( "-M has been replaced in favor of -N m,n to allow -N to be repeated.\n\n" );
            exit(1);
        }
        else if ( strcmp("-c", argv[i]) == 0 ) {
            checkres = 1;
        }
        else if ( strcmp("-c2", argv[i]) == 0 ) {
            checkres = 2;
        }
        else if ( strcmp("-l", argv[i]) == 0 ) {
            lapack = true;
        }
        else {
            printf( "invalid argument: %s\n", argv[i] );
            exit(1);
        }
    }
    if ( ntest == 0 ) {
        ntest = MAXTESTS;
        M = msize[ntest-1];
        N = nsize[ntest-1];
    }

    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    int num = 0;
    magma_err_t err;

    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", err );
      exit(-1);
    }
    
    n2     = M * N;
    min_mn = min(M, N);
    nb     = magma_get_sgetrf_nb(min_mn);

    /* Allocate memory for the matrix */
    TESTING_MALLOC_HOST(ipiv, magma_int_t, min_mn);
    TESTING_MALLOC_HOST( h_A, float, n2 );

    if ( checkres == 2 ) {
        printf("    M     N   CPU GFlop/s (sec)   GPU GFlop/s (sec)   |Ax-b|/(N*|A|*|x|)\n");
    }
    else {
        printf("    M     N   CPU GFlop/s (sec)   GPU GFlop/s (sec)   |PA-LU|/(N*|A|)\n");
    }
    printf("=========================================================================\n");
    for( i = 0; i < ntest; ++i ) {
      for(int jj=0;jj<2;jj++){
        M = msize[i];
        N = nsize[i];
        min_mn = min(M, N);
        lda    = M;
        n2     = lda*N;
        gflops = FLOPS_SGETRF( M, N ) / 1e9;

        /* =====================================================================
           Performs operation using LAPACK
           =================================================================== */
        if ( lapack ) {
            init_matrix( M, N, h_A, lda );
            
            cpu_time = magma_wtime();
            lapackf77_sgetrf(&M, &N, h_A, &lda, ipiv, &info);
            cpu_time = magma_wtime() - cpu_time;
            cpu_perf = gflops / cpu_time;

            if (info != 0)
                printf("lapackf77_sgetrf returned error %d.\n", (int) info);
        }

        /* ====================================================================
           Performs operation using MAGMA
           =================================================================== */
        init_matrix( M, N, h_A, lda );
        
        gpu_time = magma_wtime();
        ret = magma_sgetrf( M, N, h_A, lda, ipiv, &info, queue);
        gpu_time = magma_wtime() - gpu_time;

	if (info < 0)
	  printf("Argument %d of sgetrf had an illegal value.\n", -info);
	if (ret != MAGMA_SUCCESS)
	  printf("magma_sgetrf returned with error code %d\n", ret);
	
        gpu_perf = gflops / gpu_time;

        if (info != 0)
            printf("magma_sgetrf returned error %d.\n", (int) info);

        /* =====================================================================
           Check the factorization
           =================================================================== */
        printf("%5d %5d", (int) M, (int) N );
        if ( lapack ) {
            printf("   %7.2f (%7.2f)", cpu_perf, cpu_time );
        }
        else {
            printf("     ---   (  ---  )");
        }
        printf("   %7.2f (%7.2f)", gpu_perf, gpu_time );
        if ( checkres == 2 ) {
            error = get_residual( M, N, h_A, lda, ipiv );
            printf("   %8.2e\n", error );            
        }
        else if ( checkres ) {
            error = get_LU_error(M, N, h_A, lda, ipiv );
            printf("   %8.2e\n", error );
        }
        else {
            printf("     ---   \n");
        }
      }
    }

    /* Memory clean up */
    TESTING_FREE_HOST( ipiv );
    TESTING_FREE_HOST( h_A );

    /* Shutdown */
    magma_queue_destroy( queue );
    magma_finalize();
}
